from TPtoolbox.CymeInstance.HelperFuncs import FitUnimodalDist
from TPtoolbox.CymeInstance.HelperFuncs import GetReleventData
from TPtoolbox.CymeInstance.HelperFuncs import SliceTimeRange
import TPtoolbox.CymeInstance.ProfileDatabase as ProDB
import TPtoolbox.CymeInstance.Instance as Instance
from multiprocessing import Pool
import pandas as pd
import logging
import time
import os

InstanceSettings = {
    'CymeInstallationDir' : r'C:\Program Files (x86)\CYME\CYMEv8.0Rev08',
    'CymeProjectPath'     : r'C:\Users\alatif\Desktop\IEEE_13.sxst',
}

LoggerOptions =  {
        'Logging Level'          : logging.DEBUG,
        'Log to external file'   : True,
        'Display on screen'      : True,
        'Clear old log files'    : True,
}

SampleWindowSize = 5
NumMCsamples = 8760
DataSlices = 2
HourShift = 2

def runParallel(WorkerData):
    # This code is run for each parallel run (CYME instance should be created here)
    time.sleep(WorkerData)
    cymeInstance = Instance.Create(InstanceSettings, LoggerOptions, WorkerData)
    #cymeInstance.runMCsimulations(1000)

if __name__ == "__main__":
    startTime = time.time()
    # Delete old MS-Access file if it exists
    filename = os.path.join(r'C:\Users\alatif\Desktop\NetworkIterator', 'test.mdb')
    if os.path.exists(filename): os.remove(filename)
    # Create empty MSaccess database at the given location
    ProfilesDatabase = ProDB.Create(r'C:\Users\alatif\Desktop\NetworkIterator', 'test.mdb')
    # Load_data = pd.read_excel(r'C:\Users\alatif\Desktop\Load_profiles.xlsx', 0)
    # ProfilesDatabase.CreateProfiles(profilesDF=Load_data,loadPF=95,ProfileType='LOAD',ProfileResolution='Hourly')

    # Create load profile in the access database
    Load_data = pd.read_csv(r'C:\Users\alatif\Desktop\C591loadmult_Jan01toDec31.csv')
    Load_data.columns = ['c591_load']
    LDprofile = ProfilesDatabase.CreateProfiles(profilesDF=Load_data, loadPF=95, ProfileType='LOAD',
                                                ProfileResolution='1 min')
    # Create PV profile in the access database
    PV_data = pd.read_csv(r'C:\Users\alatif\Desktop\C591PVgenmult_Jan01toDec31.csv')
    PV_data.columns = ['c591_pv']
    PVprofile = ProfilesDatabase.CreateProfiles(profilesDF=PV_data, loadPF=95, ProfileType='GENERATOR',
                                                ProfileResolution='1 min')
    # Close database
    #del ProfilesDatabase

    # Get times where profiles need to be split
    SliceTimes, ProfileLens = SliceTimeRange(LDprofile, DataSlices, SampleWindowSize, HourShift)
    # Create work data for each parallel run
    MCsettings = list(zip(SliceTimes, ProfileLens, range(DataSlices),[LDprofile] * len(SliceTimes)))

    for splitTime, nDays, InstanceNumber, LDprofile in MCsettings:
        trainDF, testDF = GetReleventData(LDprofile, SplitDateTime=splitTime,
                                          MinSampleWindow=30, DaySampleWindow=SampleWindowSize)
        MCsamples = FitUnimodalDist(trainDF, testDF, MCsamples=NumMCsamples, nBins=50, split_time=splitTime, plot=True,
                            index=InstanceNumber, path=r'C:\Users\alatif\Desktop\Cyme results\Distribution fitting')
        Loads = ['Load611', 'Load634', 'Load645', 'Load646', 'Load652', 'Load671', 'Load675', 'Load692', 'L632-671-1']
        for load in Loads:
            MCsamples.columns = [load]
            LDprofile = ProfilesDatabase.CreateProfiles(profilesDF=MCsamples, loadPF=95, ProfileType='LOAD',
                                                        ProfileResolution='Hourly')
        quit()
    del ProfilesDatabase
    WorkerData = range(DataSlices)
    print('Profile sliced at : ',  str(SliceTimes))
    # Run parallel simulations
    pool = Pool(processes=DataSlices)
    pool.map(runParallel, WorkerData)
    pool.close()
    print('Total time is - ' + str(time.time() - startTime) + ' seconds.')



from cymepy.common import CORE_CYMEPY_PROJECT_FILES, DEVICES_WITH_MEMORY
from cymepy.validators import validate_settings
from cymepy.helics_interface import HELICS
from cymepy.solver import Solver
from cymepy.device_obj import DEVICE
import cymepy.logger as Logger
import os

class cymeInstance:
    def __init__(self, SettingsDict, N=None):
        self.SystemStates = []
        self.step = 0
        self.settings = validate_settings(SettingsDict, CORE_CYMEPY_PROJECT_FILES.SIMULATION_FILE)
        LoggerTag = 'CymeInstance' if N == None else 'CymeInstance_' + str(N)
        self.__Logger = Logger.getLogger(
            LoggerTag,
            LoggerOptions=SettingsDict["logger"],
            logger_path=os.path.join(self.settings["project"]["project_path"], 'logs')
        )

        import sys
        sys.path.append(SettingsDict["project"]['cyme_installation_directory'])
        self.__Logger.info('Creating Cyme instance')
        try:
            import cympy
            import cympy.rm
            import cympy.db
            cympy.app.ActivateRefresh(False)
            self.cympy = cympy
        except:
            self.__Logger.error('Cyme module not found.')
            raise Exception("Cyme module not found.")

        if self.settings["profiles"]["use_profiles"] and self.settings["profiles"]["use_internal_profile_manager"]:
            if self.settings['project']["mdb_file"]:
                conn_info = cympy.db.ConnectionInformation()
                profiles_path = os.path.join(
                    self.settings['project']["project_path"],
                    'model',
                    self.settings['project']["mdb_file"]
                )
                assert os.path.exists(profiles_path), f"The profiles file: {profiles_path} does not exist"
                conn_info.LoadProfile = cympy.db.MDBDataSource()
                conn_info.LoadProfile.Path = profiles_path
                cympy.db.Connect(conn_info)

        self.__Logger.info(cympy.version + ' created sucessfully.')
        self.projectPath = os.path.join(
            SettingsDict["project"]['project_path'],
            "model",
            SettingsDict["project"]['sxst_file']
        )

        if os.path.exists(self.projectPath):
            try:
                cympy.study.Open(self.projectPath)
            except Exception as exception:
                raise Exception(f'Unable to load the project at: {self.projectPath}\n{exception}')
            self.__Logger.info('Project loaded successfully.')
        else:
            raise Exception(f"Project path: {self.projectPath} does not exist")

        self.simObj = Solver(cympy, SettingsDict, self.__Logger)

        if self.settings["profiles"]["source_type"] == "h5":
            from cymepy.profile_manager.hooks.HDF5 import ProfileManager
            self.profile_manager = ProfileManager(self.cympy, self.simObj, self.settings, self.__Logger)
        elif self.settings["profiles"]["source_type"] == "mdb":
            from cymepy.profile_manager.hooks.ACCESS import ProfileManager
            self.profile_manager = ProfileManager(self.cympy, self.simObj, self.settings, self.__Logger)

        self.devices = {}
        if self.settings['project']["simulation_type"] == "QSTS":
            for elm_type in DEVICES_WITH_MEMORY:
                self.devices[elm_type] = self.get_devices(elm_type, DEVICES_WITH_MEMORY[elm_type])

        self.export_manager = self.get_export_manager()

        if self.settings["helics"]["cosimulation_mode"]:
            self.HI = HELICS(SettingsDict, self.cympy, self.simObj, self.__Logger)

        return

    def get_devices(self, elm_type, var_list):
        enumerator = getattr(self.cympy.enums.DeviceType, elm_type)
        devices = self.cympy.study.ListDevices(enumerator)

        device_dict = {}
        for device in devices:
            device_dict[device.DeviceNumber] = DEVICE(self.cympy, device, var_list)
        return device_dict

    def UpdateLoadProfiles(self, ProfilePath):

        return

    def runSimulation(self):
        Steps, StartTime, EndTime = self.simObj.SimulationSteps()
        self.__Logger.info('Running simulation from time {} to {} at a time increment of {} minute/s.'.format(
            StartTime, EndTime, self.settings['project']['time_step_min']
        ))
        step = 0
        incFlag = False
        while step < Steps:
            if self.step != step and step != 0:
                self.restore_states()
            incFlag = self.runStep(incFlag)
            if incFlag:
                self.export_manager.update()
                self.save_states()
                step += 1
            self.step = step
        self.export_manager.export()
        return

    def runStep(self, increment_flag):

        self.profile_manager.update()
        if self.settings['helics']['cosimulation_mode']:
            self.HI.update_subscriptions()
            pass

        if self.settings['helics']['cosimulation_mode']:
            if increment_flag:
                self.simObj.increment()
            else:
                self.simObj.resolve()
        else:
            self.simObj.increment()
        if self.settings['helics']['cosimulation_mode']:
            increment_flag, helics_time = self.HI.request_time_increment()
            self.HI.update_publications()
            return increment_flag
        else:
            return True

    def save_states(self):
        for cName, devices in self.devices.items():
            for dName, device in devices.items():
                device.save_state()
        return

    def restore_states(self):
        for cName, devices in self.devices.items():
            for dName, device in devices.items():
                device.restore_state()
        return

    def get_export_manager(self):
        if self.settings['Exports']['export_file_type'] == "csv":
            from cymepy.export_manager.hooks.csv_writer import Writer
            export_manager = Writer(self.cympy, self.simObj, self.settings, self.__Logger)
        elif self.settings['Exports']['export_file_type'] == "h5":
            from cymepy.export_manager.hooks.h5_writer import Writer
            export_manager = Writer(self.cympy, self.simObj, self.settings, self.__Logger)
        elif self.settings['Exports']['export_file_type'] == "json":
            from cymepy.export_manager.hooks.JSON_writer import Writer
            export_manager = Writer(self.cympy, self.simObj, self.settings, self.__Logger)
        return export_manager

if __name__ == "__main__":
    import toml
    Settings = toml.load(open(r"C:\Users\alatif\Desktop\Cymepy\examples\LFwithProfiles\Settings.toml"))
    instance = cymeInstance(Settings)
    instance.runSimulation()


from enum import IntEnum, Enum

DEFAULT_PROFILE_SETTINGS = {
        "multiplier": 1,
        "normalize": False,
        "interpolate": False
    }

PROPERTY_MAP = {
    "kw": "ActiveGeneration",
    "pf": "PowerFactor",
}

UNITS_MAP = {
    5: "kva",
    4: "pf",
}


CLASS_MAP = {
    14 : "SpotLoad",
}

FIELD_MAP = {
    "SpotLoad": {
        "kw": [
            "CustomerLoads.Get(",
            ").CustomerLoadModels.Get(1)"
        ],
        "kva": [
            "CustomerLoads.Get(",
            ").CustomerLoadModels.Get(1)",
        ],
        "kvar": [
            "CustomerLoads.Get(",
            ").CustomerLoadModels.Get(1)"
        ],
        "pf": [
            "CustomerLoads.Get(",
            ").CustomerLoadModels.Get(1)"
        ],
    },
    "Photovoltaic": {
            "ActiveGeneration": "GenerationModels.Get(1)",
            "PowerFactor": "GenerationModels.Get(1)"
    },
}



class PROFILE_SOURCE_TYPES(Enum):
    HDF5 = "HDF5"
    CSV = "CSV"
    MDB = "mdb"


class PROFILE_TYPES(IntEnum):
    Load = 0
    Generation = 1
    Irradiance = 2
    Temperature = 3
    Voltage = 4
    Current = 5
    EM_Price = 6
    AS_Price = 7
    WindProfile = 8

    @staticmethod
    def names():
        return list(map(lambda c: c.name, PROFILE_TYPES))

    @staticmethod
    def values():
        return list(map(lambda c: c.value, PROFILE_TYPES))



from datetime import datetime, timedelta
from cymepy.common import DATE_FORMAT
import math
import os

class Solver:

    def __init__(self, cymepy, settings, logger):
        self.Settings = settings
        self._Logger = logger
        self.cymepy = cymepy

        self._mStepRes = settings['project']['time_step_min']
        StartTimeMin = settings['project']['start_time']
        EndTimeMin = settings['project']['end_time']

        self._Time = datetime.strptime(StartTimeMin, DATE_FORMAT)
        self._StartTime = self._Time
        self._EndTime = datetime.strptime(EndTimeMin, DATE_FORMAT)

        if settings['project']["simulation_type"] == "QSTS":
            if self.Settings['profiles']["use_internal_profile_manager"]:

                self.solverObj = cymepy.sim.LoadFlowWithProfiles()
                self.solverObj.SetValue("SingleTimeMode", "Parameters.TimeParametersMode")
                self.loadflowSettings(cymepy.sim.LoadFlow())

            else:
                self.solverObj = cymepy.sim.LoadFlow()
                self.loadflowSettings(self.solverObj)
        elif settings['project']["simulation_type"] == "Static":
            self.solverObj = cymepy.sim.LoadFlow()
            self.loadflowSettings(self.solverObj)

        self._Logger.debug("Solver object created.")
        return

    def loadflowSettings(self, lf):
        lf.SetValue('VoltageDropUnbalanced', 'ParametersConfigurations[0].AnalysisMode')
        lf.SetValue(self.Settings['project']["max_iter"],
                                'ParametersConfigurations[0].MaximumIterations')
        lf.SetValue(self.Settings['project']["error_tolerance"],
                                'ParametersConfigurations[0].VoltageTolerance')
        return

    def increment(self):
        if self.Settings['project']["simulation_type"] == "QSTS":
            if self.Settings['profiles']["use_profiles"]:
                if self.Settings['profiles']["use_internal_profile_manager"]:
                    self.solverObj.SetValue(int(self._Time.timestamp()), "Parameters.SingleTime")
                    self.solverObj.Run()
                    #self._Logger.debug(f"CYME internal time: {self._Time}")
                else:
                    self.solverObj.Run()

            self._Time = self._Time + timedelta(minutes=self._mStepRes)
            self._Logger.debug(f"CYMEPY time: {self._Time}")
        elif self.Settings['project']["simulation_type"] == "Static":
            raise Exception("'increment' method cannot be used in QSTS mode")

        return

    def resolve(self):
        self.solverObj.Run()
        self._Logger.debug(f"Resolving at time: {self._Time}")

    def SimulationSteps(self):
        Minutes = (self._EndTime - self._StartTime).total_seconds() / 60.0
        Steps = math.ceil(Minutes / self._mStepRes)

        return Steps, self._StartTime, self._EndTime

    def GetTotalSeconds(self):
        return (self._Time - self._StartTime).total_seconds()

    def GetDateTime(self):
        return self._Time


import enum


DATE_FORMAT = '%Y-%m-%d %H:%M:%S.%f'
EXPORT_FILENAME = "Results"
CYMEPY_FOLDER_STRUCTURE = ["model", "logs", "exports"]

class CORE_CYMEPY_PROJECT_FILES(enum.Enum):
    SIMULATION_FILE = "Settings.toml"
    PUBLICATION_FILE = "Publications.toml"
    SUBSCRIPTION_FILE = "Subscriptions.toml"
    MAPPING_FILE = "Mapping.toml"
    PROFILES = "Profiles.hdf5"
    EXPORT_FILE = "Exports.toml"


DEVICES_WITH_MEMORY = {
    'BESS': [],
    'Battery': [],
    'Regulator': ["RegTapA", "RegTapB", "RegTapC"],
    'ShuntCapacitor': ["CapStatus"],
}




"""
CLI to run a PyDSS project
"""

from cymepy.common import CORE_CYMEPY_PROJECT_FILES
from cymepy.utils.utils import readToml
from cymepy.cymepy import cymeInstance
import logging
import click

import os

logger = logging.getLogger(__name__)

@click.argument(
    "project-path",
)

@click.option(
    "-s", "--simulations-file",
    required=False,
    default = CORE_CYMEPY_PROJECT_FILES.SIMULATION_FILE.value,
    show_default=True,
    help="scenario toml file to run (over rides default)",
)

@click.command()

def run(project_path, simulations_file):
    file_path = os.path.join(project_path, simulations_file)
    settings = readToml(file_path)
    instance = cymeInstance(settings)
    instance.runSimulation()
    return




from cymepy.common import CORE_CYMEPY_PROJECT_FILES
import toml
import abc
import os
import re

import logging
logger = logging.getLogger(__name__)

class ExportManager(abc.ABC):
    def __init__(self, sim_instance, solver, options, logger, **kwargs):
        self._Logger = logger
        self.cympy = sim_instance
        self.settings = options
        self.solver = solver
        pubpath = os.path.join(
            self.settings["project"]['project_path'],
            CORE_CYMEPY_PROJECT_FILES.EXPORT_FILE.value
        )
        exportDict = toml.load(open(pubpath, "r"))

        self.validTypes = []
        for k, v in self.cympy.enums.DeviceType.__dict__.items():
            if isinstance(v, int):
                self.validTypes.append(k)

        self.Exports = {}
        for cName, exportList in exportDict.items():
            for exportInfo in exportList:
                if cName not in self.validTypes:
                    raise Exception(f"{cName} is not a valid CYME device type. "
                                    f"For valid device type, see cympy.enums.DeviceType")
                else:
                    devType = getattr(self.cympy.enums.DeviceType, cName)
                    devices = self.cympy.study.ListDevices(devType)
                    if devices:
                        for device in devices:
                            eName = device.DeviceNumber
                            if exportInfo["regex_filter"]:
                                pattern = re.compile(exportInfo["regex_filter"])
                                matches = pattern.search(eName)
                                if matches:

                                    self.create_export_list(exportInfo, device, cName, eName)
                            else:
                                self.create_export_list(exportInfo, device, cName, eName)
                    else:
                        self._Logger.warn(f"Model of type {cName} not found in the distribution model")

        return

    def create_export_list(self, pubInfo, device, cName, eName):
        if cName not in self.Exports:
            self.Exports[cName] = {}
        if eName not in self.Exports[cName]:
            self.Exports[cName][eName] = {
                "device": device,
                "regex": pubInfo['regex_filter'],
                "properties": pubInfo['properties'],
            }
        return

    @abc.abstractmethod
    def update(self):
        time = self.solver.GetDateTime().isoformat()
        results = {time: {}}
        for cName, cInfo in self.Exports.items():
            if cName not in results[time]:
                results[time][cName] = {}
            for eName, einfo in cInfo.items():
                if eName not in results[time][cName]:
                    results[time][cName][eName] = {}
                for ppty in einfo['properties']:
                    if ppty not in results[time][cName][eName]:
                        results[time][cName][eName][ppty] = None
                    device = einfo['device']
                    keyword = self.cympy.app.GetKeyword(ppty)
                    if keyword:
                        res = self.cympy.study.QueryInfoDevice(ppty, device.DeviceNumber, device.DeviceType)
                    else:
                        res = device.GetValue(ppty)
                    results[time][cName][eName][ppty] = res
        return results




from cymepy.export_manager.base_definations import ExportManager
from cymepy.common import EXPORT_FILENAME
import json
import os

class Writer(ExportManager):
    def __init__(self, sim_instance, solver, options, logger, **kwargs):
        super(Writer, self).__init__(sim_instance, solver, options, logger, **kwargs)
        self.results = []
        self.path = os.path.join(
            self.settings["project"]['project_path'],
            'exports',
            f"{EXPORT_FILENAME}.json"
        )
        pass

    def update(self):
        results = super().update()
        self.results.append(results)
        return

    def export(self):
        with open(self.path, "w") as write_file:
            json.dump(self.results, write_file, indent=4, sort_keys=True)



import toml
import os

def readToml(file_path):
    if os.path.exists(file_path):
        data = toml.load(open(file_path, "r"))
        return data
    else:
        raise Exception(f"File path does not exist: {file_path}")

def writeToml(file_path, data_dict):
    data = toml.dump(data_dict, open(file_path, "w"))


from cymepy.export_manager.base_definations import ExportManager
from cymepy.common import EXPORT_FILENAME
import numpy as np
import h5py
import os


class Writer(ExportManager):
    def __init__(self, sim_instance, solver, options, logger, **kwargs):
        super(Writer, self).__init__(sim_instance, solver, options, logger, **kwargs)
        self.i = 0
        self.columnLength,_ , _ = solver.SimulationSteps()
        self.logger = logger
        self.basepath = os.path.join(
            self.settings["project"]['project_path'],
            'exports',
            f"{EXPORT_FILENAME}.h5"
        )

        self.logger.info("Creating new h5 store")
        self.store = h5py.File(self.basepath, "w")
        self.Timestamp = self.store.create_dataset(
            "Timestamp",
            shape=(self.columnLength,),
            maxshape=(None,),
            chunks=True,
            compression="gzip",
            compression_opts=4,
            shuffle=True,
            dtype="S30",
        )
        self.store_groups = {}
        for cName, eInfo in self.Exports.items():
            cGroup = self.store.create_group(cName)
            if cName not in self.store_groups:
                self.store_groups[cName] = {}
            for eName, pInfo in eInfo.items():
                eGroup = cGroup.create_group(eName)
                if eName not in self.store_groups[cName]:
                    self.store_groups[cName][eName] = {}
                for ppty in pInfo["properties"]:
                    self.store_groups[cName][eName][ppty] = eGroup.create_dataset(
                        str(ppty),
                        shape=(self.columnLength, ),
                        maxshape=(None, ),
                        chunks=True,
                        compression="gzip",
                        compression_opts=4,
                        shuffle=True
                )
        return

    def update(self):
        results = super().update()
        for timestamp, cInfo in results.items():
            self.Timestamp[self.i] = np.string_(timestamp)
            for cName, eInfo in cInfo.items():
                for eName, pptyInfo in eInfo.items():
                    for ppty, value in pptyInfo.items():
                        try:
                            self.store_groups[cName][eName][ppty][self.i] = float(value)
                        except:
                            self.store_groups[cName][eName][ppty][self.i] = np.NaN
        self.i += 1
        return

    def export(self):
        self.store.flush()
        self.store.close()
        return




class DEVICE:

    def __init__(self, cymepy, device, memory_valiables):
        self.cymepy = cymepy
        self.device = device
        self.memory_valiables = memory_valiables
        return

    def save_state(self):
        self.states = {}
        for param in self.memory_valiables:
            self.states[param] = self.cymepy.study.QueryInfoDevice(param, self.device.DeviceNumber, self.device.DeviceType)

        return

    def restore_state(self):
        if hasattr(self, 'states'):
            for param, val in self.states.items():
                print(self.device, param, val)
                #self.device.SetValue(val, "TapA")

                # self.cymepy.study.SetValueDevice(
                #     val,
                #     param,
                #     self.device.DeviceNumber,
                #     self.device.DeviceType
                # )
        return


from cymepy.common import CORE_CYMEPY_PROJECT_FILES
from cymepy.validators import validate_settings
import helics as h
import toml
import os
import re

class HELICS:

    n_states = 5
    init_state = 1

    type_info = {
        'complex': 'complex ',
        'text': 'string',
        'number': 'int64',
        'real': 'double',
        '': 'double',
    }

    def __init__(self, settings, cymepy, Solver, logger):
        self.__Logger = logger
        self.cympy = cymepy
        self.settings = settings
        self.Solver = Solver
        lf = cymepy.sim.LoadFlow()
        lf.Run()

        self.itr = 0
        self.c_seconds = 0
        self.c_seconds_old = -1

        self.validTypes = []
        for k, v in self.cympy.enums.DeviceType.__dict__.items():
            if isinstance(v, int):
                self.validTypes.append(k)

        self.create_helics_federate()
        self.registerPublications()
        self.registerSubscriptions()

        h.helicsFederateEnterExecutingModeIterative(
            self.cymeFederate,
            h.helics_iteration_request_iterate_if_needed
        )
        logger.info('Entered HELICS execution mode')

    def registerSubscriptions(self):
        self.__Logger.info("Creating subscriptions")
        subpath = os.path.join(
            self.settings["project"]['project_path'],
            CORE_CYMEPY_PROJECT_FILES.SUBSCRIPTION_FILE.value
        )

        subscriptionDict = toml.load(open(subpath, "r"))
        subscriptionDict = validate_settings(subscriptionDict, CORE_CYMEPY_PROJECT_FILES.SUBSCRIPTION_FILE)
        self.Subscriptions = {}
        for elmCN, subInfoList in subscriptionDict.items():
            for subInfo in subInfoList:
                cName, eName = elmCN.split(".")
                if cName not in self.validTypes:
                    raise Exception(f"{cName} is not a valid CYME device type. "
                                    f"For valid device type, see cympy.enums.DeviceType")
                else:
                    devType = getattr(self.cympy.enums.DeviceType, cName)
                    devices = self.cympy.study.ListDevices(devType)
                    print(devices)
                    found = False
                    fDevice = None
                    for device in devices:
                        if eName == device.DeviceNumber:
                            found = True
                            fDevice = device
                            break
                    if not found:
                        raise Exception(f"Element {eName} for class {cName} not found. See subscriptions")
                    else:

                        found, pUnits, pDefaultValue, pType = self.isProperty(fDevice, subInfo["property"])
                        if not found:
                            raise Exception("{} is not a valid property for {}.\nValid options are:\n{}".format(
                                subInfo['property'],
                                cName,
                                self.listProperties(cName)
                            ))

                        subname = subInfo["subscription"]
                        if subname not in self.Subscriptions:
                            self.Subscriptions[subname] = []
                        self.Subscriptions[subname].append({
                            "elementObj": fDevice,
                            "subscriptionObj": h.helicsFederateRegisterSubscription(
                                self.cymeFederate,
                                subInfo["subscription"],
                                pUnits
                            ),
                            "class": cName,
                            "name": eName,
                            "property": subInfo["property"],
                            "unit": pUnits,
                            "dType": pType,
                            "dStates": [self.init_state] * self.n_states,
                            "mult" : subInfo["multiplier"]
                        })
                        self.__Logger.debug(
                            f"Object {cName}.{eName}'s {subInfo['property']} property has subscribed to {subInfo['subscription']}.")

        return

    def listProperties(self, cName):
        properties = self.cympy.dm.Describe(cName)
        string = ""
        for i, p in enumerate(properties):
            string += f"{i}. {p.Name}\n"
        return string

    def create_helics_federate(self):
        self.fedName = self.settings['helics']['federate_name']
        self.fedinfo = h.helicsCreateFederateInfo()
        h.helicsFederateInfoSetCoreName(self.fedinfo, self.settings['helics']['federate_name'])
        h.helicsFederateInfoSetCoreTypeFromString(self.fedinfo, self.settings['helics']['core_type'])
        h.helicsFederateInfoSetCoreInitString(self.fedinfo, f"--federates=1")
        h.helicsFederateInfoSetBroker(self.fedinfo, self.settings['helics']['broker'])
        h.helicsFederateInfoSetBrokerPort(self.fedinfo ,self.settings['helics']['broker_port'])
        h.helicsFederateInfoSetTimeProperty(self.fedinfo, h.helics_property_time_delta,
                                                 self.settings['helics']['time_delta'])
        h.helicsFederateInfoSetIntegerProperty(self.fedinfo, h.helics_property_int_log_level,
                                                    self.settings['helics']['helics_logging_level'])

        #h.helicsFederateInfoSetFlagOption(self.fedinfo, h.helics_flag_uninterruptible, True)
        h.helicsFederateInfoSetIntegerProperty(self.fedinfo, h.helics_property_int_max_iterations,
                                                    self.settings["helics"]["max_coiter"])
        self.cymeFederate = h.helicsCreateValueFederate(self.settings['helics']['federate_name'], self.fedinfo)

        return

    def registerPublications(self):
        pubpath = os.path.join(
            self.settings["project"]['project_path'],
            CORE_CYMEPY_PROJECT_FILES.PUBLICATION_FILE.value
        )

        publicationDict = toml.load(open(pubpath, "r"))
        publicationDict = validate_settings(publicationDict, CORE_CYMEPY_PROJECT_FILES.PUBLICATION_FILE)

        self.Publications = {}
        for cName, pubInfoList in publicationDict.items():
            for pubInfo in pubInfoList:
                if cName not in self.validTypes:
                    raise Exception(f"{cName} is not a valid CYME device type. "
                                    f"For valid device type, see cympy.enums.DeviceType")
                else:
                    devType = getattr(self.cympy.enums.DeviceType, cName)
                    devices = self.cympy.study.ListDevices(devType)
                    if devices:
                        for device in devices:
                            eName = device.DeviceNumber
                            if pubInfo["regex_filter"]:
                                pattern = re.compile(pubInfo["regex_filter"])
                                matches = pattern.search(eName)
                                if matches:
                                    self.create_publication(pubInfo, device, cName, eName)
                            else:
                                self.create_publication(pubInfo, device, cName, eName)
                    else:
                        self.__Logger.warn(f"Model of type {cName} not found in the distribution model")
        return

    def isProperty(self,device, property):
        properties = self.cympy.dm.Describe(device.GetObjType())
        found = False
        for p in properties:
            if p.Name == property:
                found = True
                return found, p.Units, p.DefaultValue, p.Type

        return found, None, None, None

    def create_publication(self, pubInfo, device, cName, eName):
        self.__Logger.info("Creating publications")
        for propertyX in pubInfo["properties"]:
            keyword = self.cympy.app.GetKeyword(propertyX)
            if keyword is None:
                found, units, defaultValue, cType = self.isProperty(device, propertyX)
                if not found:
                    raise Exception(f"{propertyX} is neither a valid keyword not a valid property of object type {cName}")
                res = device.GetValue(propertyX)
            else:
                res = self.cympy.study.QueryInfoDevice(propertyX, device.DeviceNumber, device.DeviceType)
            pubname = f"{self.fedName}.{cName}.{eName}.{propertyX}"
            T = self.type_info[keyword.Type.lower()] if keyword is not None else self.type_info[cType]
            if res:
                self.Publications[pubname] = {
                    "elementObj": device,
                    "publicationObj": h.helicsFederateRegisterGlobalTypePublication(
                        self.cymeFederate,
                        pubname,
                        T,
                        keyword.Unit if keyword is not None else units
                    ),
                    "class": cName,
                    "name": eName,
                    "property": propertyX,
                    "unit": keyword.Unit if keyword is not None else units,
                    "dType": keyword.Type if keyword is not None else cType,
                    "isKeyword": True if keyword is not None else False
                }
                self.__Logger.debug(
                    f"Publication {pubname} of type {T} registered successfully.")
            else:
                self.__Logger.warning(f"Property {propertyX} for publication {pubname} not valid")
        return

    def update_publications(self):
        for pubName, pubInfo in self.Publications.items():
            device = pubInfo["elementObj"]
            pub = pubInfo["publicationObj"]
            if pubInfo["isKeyword"]:
                res = self.cympy.study.QueryInfoDevice(pubInfo["property"], device.DeviceNumber, device.DeviceType)
            else:
                res = pubInfo["elementObj"].GetValue(pubInfo["property"])

            print(pubInfo["property"], res, type(res), pubInfo["dType"], pubInfo["unit"])

            if pubInfo["dType"].lower() == "complex":
                h.helicsPublicationPublishComplex(pub, complex(res).real, complex(res).imag)
            elif pubInfo["dType"].lower() == 'real':
                h.helicsPublicationPublishDouble(pub, float(res))
            elif pubInfo["dType"].lower() == 'text':
                h.helicsPublicationPublishString(pub, str(res))
            elif pubInfo["dType"].lower() == "number":
                h.helicsPublicationPublishInteger(pub, int(res))

        return

    def update_subscriptions(self):
        for subName, subList in self.Subscriptions.items():
            for subInfo in subList:
                sub = subInfo["subscriptionObj"]
                if subInfo["dType"].lower() == "complex":
                    value = h.helicsInputGetComplex(sub)
                elif subInfo["dType"].lower() == 'real':
                    value = h.helicsInputGetDouble(sub)
                elif subInfo["dType"].lower() == 'text':
                    value = h.helicsInputGetString(sub)
                elif subInfo["dType"].lower() == "number":
                    value = h.helicsInputGetInteger(sub)
                else:
                    value = h.helicsInputGetDouble(sub)

                if value:
                    value = value * subInfo["mult"]
                    X1 = subInfo["elementObj"].GetValue(subInfo["property"])

                    if value < 10000.0 and value > -100000.0:
                        subInfo["elementObj"].SetValue(value, subInfo["property"])
                        #X2 = subInfo["elementObj"].GetValue(subInfo["property"])
                        self.__Logger.debug(f"{subInfo['class']}.{subInfo['name']}.{subInfo['property']} updated to {value}")
                    if self.settings['helics']['coiter_mode']:
                        if self.c_seconds != self.c_seconds_old:
                            subInfo["dStates"] = [self.init_state] * self.n_states
                        else:
                            subInfo["dStates"].insert(0, subInfo["dStates"].pop())
                        subInfo["dStates"][0] = value
            #self.cympy.study.Save()
        return

    def request_time_increment(self):
        error = sum([abs(y["dStates"][0] - y["dStates"][1]) for k, x in self.Subscriptions.items() for y in x])
        r_seconds = self.Solver.GetTotalSeconds()
        if not self.settings['helics']['coiter_mode']:
            while self.c_seconds < r_seconds:
                self.c_seconds = h.helicsFederateRequestTime(self.cymeFederate, r_seconds)
            self.__Logger.info('Time requested: {} - time granted: {} '.format(r_seconds, self.c_seconds))
            return True, self.c_seconds
        else:
            self.c_seconds, iteration_state = h.helicsFederateRequestTimeIterative(
                self.cymeFederate,
                r_seconds,
                h.helics_iteration_request_iterate_if_needed
            )
            self.__Logger.info('Time requested: {} - time granted: {} error: {} it: {}'.format(
                r_seconds, self.c_seconds, error, self.itr))
            if error > -1 and self.itr < self.settings['helics']["max_coiter"]:
                self.itr += 1
                return False, self.c_seconds
            else:
                self.itr = 0
                return True, self.c_seconds

    # def __del__(self):
    #     h.helicsFederateFinalize(self.cymeFederate)
    #     #state = h.helicsFederateGetState(self.cymeFederate)
    #     h.helicsFederateInfoFree(self.fedinfo)
    #     h.helicsFederateFree(self.cymeFederate)
    #     self.__Logger.info('HELICS federate for PyDSS destroyed.')


from cymepy.profile_manager.common import FIELD_MAP, PROPERTY_MAP
from cymepy.common import CORE_CYMEPY_PROJECT_FILES
import numpy as np
import toml
import abc
import os

import logging
logger = logging.getLogger(__name__)

class BaseProfileManager(abc.ABC):
    def __init__(self, sim_instance, solver, options, logger, **kwargs):
        self.sim_instance = sim_instance
        self.options = options
        self.solver = solver
        self.logger = logger
        self.kwargs = kwargs
        if not options["profiles"]["is_relavtive_path"]:
            self.basepath = options["profiles"]["source"]
        else:
            self.basepath = os.path.join(
                options["project"]["project_path"], options["profiles"]["source"]
            )
        self.mapping_file = os.path.join(
            options["project"]["project_path"], CORE_CYMEPY_PROJECT_FILES.MAPPING_FILE.value
        )
        self.Profiles = {}
        self.mapping = toml.load(open(self.mapping_file , "r"))
        self.sTime = None
        self.eTime = None
        self.simRes = None
        pass

    @abc.abstractmethod
    def setup_profiles(self):
        pass

    @abc.abstractmethod
    def update(self):
        pass


class BaseProfile(abc.ABC):
    def __init__(self, sim_instance, dataset, devices, solver, mapping_dict, logger, **kwargs):
        self.sim_instance = sim_instance
        self.mapping_dict = mapping_dict
        self.dataset = dataset
        self.devices = devices
        self.logger = logger
        self.solver = solver
        self.kwargs = kwargs

        self.isLoad = [
            self.sim_instance.enums.DeviceType.DCLoad,
            self.sim_instance.enums.DeviceType.SpotLoad,
            self.sim_instance.enums.DeviceType.DistributedLoad,
        ]

        self.isPV = self.sim_instance.enums.DeviceType.Photovoltaic

        pass

    @abc.abstractmethod
    def update_profile_settings(self):
        pass

    @abc.abstractmethod
    def update(self):
        pass

    def write(self, value, value1):
        for objName, obj in self.devices.items():
            class_name, element_name = objName.split(".")
            if self.valueSettings[objName]['interpolate']:
                value = value1
            mult = self.valueSettings[objName]['multiplier']
            if self.valueSettings[objName]['normalize']:
                valueF = value / self.attrs["max"] * mult
            else:
                valueF = value * mult
            if isinstance(self.attrs["units"], np.ndarray):
                unit = self.attrs["units"][0].decode()
            else:
                unit = self.attrs["units"].decode()

            phases = self.sim_instance.study.QueryInfoDevice("LoadPhase", obj.DeviceNumber, obj.DeviceType)
            phasesType = self.sim_instance.study.QueryInfoDevice("LoadPhaseType", obj.DeviceNumber, obj.DeviceType)
            if phasesType == 'ByPhase':
                if phases:
                    phases = [Ph for Ph in phases]
                else:
                    phases = ["A", "B", "C"]
                    self.logger.warning(f"No phase info returned for element: {class_name}.{element_name}. Defaulting to ABC configuration")
                loadMult = 1 / len(phases)
            else:
                phases = [phases]
                loadMult = 1

            if obj.DeviceType in self.isLoad:
                ppty = FIELD_MAP[class_name][unit]
                key = f"{ppty[0]}{obj.DeviceNumber}{ppty[1]}.LoadValueType"
                loadtype = obj.GetValue(key).lower()

                if unit not in loadtype.split("_"):
                    self.logger.warning(
                        f"{class_name}.{element_name} is of type {loadtype}. A {unit} profile has been attached. Value could not be updated.")
                else:
                    load = self.sim_instance.study.GetLoad(obj.DeviceNumber, self.sim_instance.enums.LoadType.Spot)
                    for ph in phases:
                        en = getattr(self.sim_instance.enums.Phase, ph)
                        load.SetValue(valueF * loadMult, f"LoadValue.{unit.upper()}", obj.DeviceNumber, en, "DEFAULT")
                        kw = load.GetValue(f"LoadValue.{unit.upper()}", obj.DeviceNumber, en, "DEFAULT")

            if obj.DeviceType is self.sim_instance.enums.DeviceType.Photovoltaic:
                if unit in PROPERTY_MAP:
                    relevant_ppty = PROPERTY_MAP[unit]
                    key = FIELD_MAP[class_name][relevant_ppty]
                    obj.SetValue(valueF * loadMult, f"{key}.{relevant_ppty}")
                else:
                    self.logger.warning(f"Units defined for {unit} are  invalid for obj type {class_name}")

            if obj.DeviceType is self.sim_instance.enums.DeviceType.Source:
                Phases = ["A", "B", "C"]
                V = valueF * loadMult
                if unit.lower() == "kvll":
                    V = valueF * loadMult / 1.732
                elif unit.lower() == "kvln":
                    V = valueF * loadMult
                for pha in Phases:
                    ppty = f"OperatingVoltage{pha}"
                    obj.SetValue(V, ppty)
                    logger.info(f"Property {ppty} updated to {V} for {obj.DeviceType}.{obj.DeviceNumber} ")
        return valueF



from cymepy.common import DATE_FORMAT, CORE_CYMEPY_PROJECT_FILES, CYMEPY_FOLDER_STRUCTURE
from pydantic import BaseModel, validator, conint, confloat
from cymepy.profile_manager.common import PROFILE_SOURCE_TYPES
from cymepy.utils.utils import readToml
from datetime import datetime
from typing import List, Dict
from enum import Enum
import os
import re

class SIMULATION_TYPE(Enum):
    Static = "Static"
    QSTS = 'QSTS'

class LOGGING_LEVEL(Enum):
    WARNING = "WARNING"
    ERROR = "ERROR"
    DEBUG = "DEBUG"
    INFO = "INFO"

class HELICS_CORE_TYPE(Enum):
    Interprocess = "Interprocess"
    ZMQ_SS = "ZMQ_SS"
    TCP_SS = "TCP_SS"
    Test = "Test"
    ZMQ = "ZMQ"
    UDP = "UDP"
    TCP = "TCP"
    MPI = "MPI"

class Project_Settings(BaseModel):
    start_time: str
    end_time: str
    time_step_min: confloat(gt=0)
    max_iter: conint(ge=1, le=1000)
    error_tolerance: confloat(gt=0)
    simulation_type: SIMULATION_TYPE
    project_path: str
    mdb_file: str
    sxst_file: str
    cyme_installation_directory: str

    @validator('start_time')
    def validate_start_time(cls, v):
        try:
            Date = datetime.strptime(v, DATE_FORMAT)
        except:
            raise ValueError("'start_time' should be a datetime string with format {}".format( DATE_FORMAT  ))
        pass

    @validator('end_time')
    def validate_end_time(cls, v):
        try:
            Date = datetime.strptime(v, DATE_FORMAT)
        except:
            raise ValueError("'end_time' should be a datetime string with format {}".format(DATE_FORMAT))
        pass

    @validator('cyme_installation_directory')
    def validate_cyme_installation_directory(cls, v):
        if not os.path.exists(v):
            raise Exception(f"Project path: {v} does not exist")

    @validator('project_path')
    def validate_project_path(cls, v):
        if not v:
            raise Exception(f"Please provide a valid project path")
        if not os.path.exists(v):
            raise Exception(f"Cyme installation path: {v} does not exist")
        else:
            for fldr in CYMEPY_FOLDER_STRUCTURE:
                if not os.path.exists(os.path.join(v, fldr)):
                    raise Exception(f"A valid cymepy project should have a '{fldr}' folder")
            for file in CORE_CYMEPY_PROJECT_FILES:
                if not os.path.exists(os.path.join(v, file.value)):
                    raise Exception(f"A valid cymepy project should have a '{file.value}' file in the base path")

class Helics_Settings(BaseModel):
    cosimulation_mode: bool
    coiter_mode: bool
    coiter_error_tolerance: confloat(gt=0)
    max_coiter: conint(ge=1, le=1000)
    broker: str
    broker_port: conint(ge=1, le=65535)
    federate_name: str
    time_delta: confloat(gt=0.0)
    core_type: HELICS_CORE_TYPE
    uninterruptible: bool
    helics_logging_level: conint(ge=1, le=7)

class Logger_Settings(BaseModel):
    logging_level: LOGGING_LEVEL
    log_to_external_file: bool
    display_on_screen: bool
    clear_old_log_file: bool
    pre_configured_logging: bool

class Profile_Settings(BaseModel):
    use_profiles: bool
    use_internal_profile_manager: bool
    source_type = PROFILE_SOURCE_TYPES
    source = str
    is_relavtive_path = bool

class Cymepy_Settings(BaseModel):
    project: Project_Settings
    helics: Helics_Settings
    logger: Logger_Settings
    profiles: Profile_Settings

class PUBLICATION(BaseModel):
    regex_filter: str
    properties: List[str]

    @validator('regex_filter')
    def validate_regex_filter(cls, v):
        try:
            re.compile(v)
        except re.error as e:
            raise Exception(f"'{v}' is not a valid regular expression. Check https://docs.python.org/3/library/re.html")

class PUBLICATIONS(BaseModel):
    publication_dict: Dict[str, List[PUBLICATION]]

class SUBSCRIPTION(BaseModel):
    property: str
    subscription: str
    multiplier: float

class SUBSCRIPTIONS(BaseModel):
    subscription_dict: Dict[str, List[SUBSCRIPTION]]

def validate_settings(project_settings, sType):
    ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
    settings_path = os.path.join(ROOT_DIR, "defaults", sType.value)
    default_settings = readToml(settings_path)
    default_settings.update(project_settings)
    if sType == CORE_CYMEPY_PROJECT_FILES.SIMULATION_FILE:
        Cymepy_Settings(**default_settings)
    elif sType == CORE_CYMEPY_PROJECT_FILES.PUBLICATION_FILE:
        PUBLICATIONS(publication_dict=default_settings)
    elif sType == CORE_CYMEPY_PROJECT_FILES.SUBSCRIPTION_FILE:
        SUBSCRIPTIONS(subscription_dict=default_settings)
    return default_settings





import logging
import os

def getLogger(name, LoggerOptions=None, logger_path=None):
    if LoggerOptions['clear_old_log_file']:
        test = os.listdir(os.getcwd())
        for item in test:
            if item.endswith(name + ".log"):
                os.remove(item)
    formatter = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(module)s - %(message)s')

    logger = logging.getLogger(name)
    logger.setLevel(LoggerOptions['logging_level'])
    if LoggerOptions['display_on_screen']:
        handler1 = logging.StreamHandler()
        handler1.setFormatter(formatter)
        logger.addHandler(handler1)
    if LoggerOptions['log_to_external_file']:
        if logger_path:
            handler2 = logging.FileHandler(filename= os.path.join(logger_path, name + '.log'))
        else:
            handler2 = logging.FileHandler(filename= name + '.log')
        handler2.setFormatter(formatter)
        logger.addHandler(handler2)
    return logger



"""Main CLI command for PyDSS."""

import logging
import click

from cymepy.cli.run import run


logger = logging.getLogger(__name__)


@click.group()
def cli():
    """PyDSS commands"""


cli.add_command(run)



#!/usr/bin/env python
import io
import os
import re

try:
    from setuptools import setup , find_packages
except ImportError:
    from distutils.core import setup

with open('requirements.txt') as f:
    all_lines = f.read().splitlines()
    print(all_lines)
requirements = [x for x in all_lines if "git+" not in x]
dependencies = [x for x in all_lines if "git+" in x]


print("cymepy installation requirements:")
for i, req in enumerate(requirements):
    print(f"{i}. {req}")

print("cymepy installation dependencies:")
for i, req in enumerate(dependencies):
    print(f"{i}. {req}")

# Read the version from the __init__.py file without importing it
def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        ver =  version_match.group(1)
        return ".".join(ver.split(".")[:-1])
    raise RuntimeError("Unable to find version string.")

print(find_version("cymepy", "__init__.py"),)

setup(
    name='cymepy',
    version=find_version("cymepy", "__init__.py"),
    description='Helice interface for CYME',
    author='Aadil Latif',
    author_email='Aadil.Latif@nrel.gov',
    url='https://github.com/GMLC-TDC/cymepy',
    packages=find_packages(),
    install_requires=requirements,
    include_package_data=True,
    #dependency_links=dependencies,
    package_data={'cymepy': ['*.toml']},
        entry_points={
            "console_scripts": [
                "cymepy=cymepy.cli.cymepy:cli",
            ],
        },
    license='BSD 3 clause',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
    ],
)



from cymepy.profile_manager.common import PROFILE_TYPES, DEFAULT_PROFILE_SETTINGS
from cymepy.profile_manager.base_definations import BaseProfileManager, BaseProfile
from cymepy.common import DATE_FORMAT
import datetime
import h5py
import copy
import os

class ProfileManager(BaseProfileManager):
    def __init__(self, sim_instance, solver, options, logger, **kwargs):

        super(ProfileManager, self).__init__(sim_instance, solver, options, logger, **kwargs)
        if os.path.exists(self.basepath):
            self.logger.info("Loading existing h5 store")
            self.store = h5py.File(self.basepath, "r+")
        else:
            self.logger.info("Creating new h5 store")
            self.store = h5py.File(self.basepath, "w")
            for profileGroup in PROFILE_TYPES.names():
                self.store.create_group(profileGroup)
        self.setup_profiles()
        pass

    def setup_profiles(self):

        for group, profileMap in self.mapping.items():
            if group in self.store:
                grp = self.store[group]
                for profileName, mappingDict in profileMap.items():
                    if profileName in grp:
                        devices = self.get_devices(mappingDict)
                        if devices:
                            self.Profiles[f"{group}/{profileName}"] = Profile(
                                self.sim_instance,
                                grp[profileName],
                                devices,
                                self.solver,
                                mappingDict,
                                self.logger,
                                **self.kwargs
                            )
                            self.logger.debug("Profile {}\{} attached to devices {}".format(
                                group, profileName, ", ".join(list(devices.keys()))
                            ))
                    else:
                        self.logger.warning("Group {} \ data set {} not found in the h5 store".format(
                            group, profileName
                        ))
            else:
                self.logger.warning("Group {} not found in the h5 store".format(group))
        return
        pass

    def get_devices(self, mappingDict):
        devicesObjects = {}
        for x in mappingDict:
            cName, eName = x['object'].split(".")
            devType = getattr(self.sim_instance.enums.DeviceType, cName)
            devices = self.sim_instance.study.ListDevices(devType)
            found = False
            for device in devices:
                if eName == device.DeviceNumber:

                    devicesObjects[x['object']] = device
                    found = True
                    break
                elif device.DeviceNumber.lower() == cName.lower():
                    devicesObjects[x['object']] = device
                    found = True
                    break

            if not found:
                self.logger.warning(
                    f"Device {eName} of class {cName} defined in the mapping file not found in the loaded model"
                )

        return devicesObjects

    def update(self):
        results = {}
        for profileaName, profileObj in self.Profiles.items():
            result = profileObj.update()
            results[profileaName] = result
        return results

class Profile(BaseProfile):
    def __init__(self, sim_instance, dataset, devices, solver, mapping_dict, logger, **kwargs):
        super(Profile, self).__init__(sim_instance, dataset, devices, solver, mapping_dict, logger, **kwargs)
        self.valueSettings = {x['object']: {**DEFAULT_PROFILE_SETTINGS, **x} for x in mapping_dict}
        self.attrs = self.dataset.attrs
        self.update_profile_settings()


        pass

    def update_profile_settings(self):
        self.sTime = datetime.datetime.strptime(self.attrs["sTime"].decode(), DATE_FORMAT)
        self.eTime = datetime.datetime.strptime(self.attrs["eTime"].decode(), DATE_FORMAT)
        self.simRes, _, _ = self.solver.SimulationSteps()
        self.Time = copy.deepcopy(self.solver.GetDateTime())

    #
    # def write(self, value1):
    #     super(Profile, self).write(value1)

    def update(self):
        self.Time = copy.deepcopy(self.solver.GetDateTime())
        if self.Time < self.sTime or self.Time > self.eTime:
            value = 0
            value1 = 0
            valueF = 0
        else:
            dT = (self.Time - self.sTime).total_seconds()
            n = int(dT / self.attrs["resTime"])
            value = self.dataset[n]
            dT2 = (self.Time - (
                        self.sTime + datetime.timedelta(seconds=int(n * self.attrs["resTime"])))).total_seconds()
            value1 = self.dataset[n] + (self.dataset[n + 1] - self.dataset[n]) * dT2 / self.attrs["resTime"]
            valueF = self.write(value, value1)

        return valueF


from cymepy.profile_manager.common import DEFAULT_PROFILE_SETTINGS, UNITS_MAP, CLASS_MAP
from cymepy.profile_manager.base_definations import BaseProfileManager, BaseProfile
import pandas as pd
import pyodbc
import copy

class ProfileManager(BaseProfileManager):

    time_mapping = {
        "1MINUTE": 1,
        "5MINUTES": 5,
        "10MINUTES": 10,
        "15MINUTES": 15,
        "30MINUTES": 30,
        "1HOUR": 60,
    }

    def __init__(self, sim_instance, solver, options, logger, **kwargs):
        super(ProfileManager, self).__init__(sim_instance, solver, options, logger, **kwargs)
        self.DRV = options["profiles"]["settings"]["driver"]
        self.PWD = options["profiles"]["settings"]["password"]
        self.freq = {}
        if self.PWD:
            self.con = pyodbc.connect('DRIVER={};DBQ={};PWD={}'.format(self.DRV, self.basepath, self.PWD))
        else:
            self.con = pyodbc.connect('DRIVER={};DBQ={}'.format(self.DRV, self.basepath))
        self.cur = self.con.cursor()
        self.setup_profiles()
        pass

    def setup_profiles(self):
        data = {}
        for dTtype, tblname in self.options["profiles"]["tables"].items():
            self.freq[dTtype] = self.time_mapping[tblname.split("_")[2]]
            SQL = f'SELECT * FROM {tblname};'  # your query goes here
            pData= pd.read_sql(SQL, self.con)
            year = min(set(pData.ProfileYear.tolist()))
            hour = min(set(pData.YearIntervalNumber.tolist())) - 1
            IDS = set(pData.ID.tolist())
            elements = {}
            for id in IDS:
                elements[id] = {}
                profileData = pData[pData["ID"] == id]
                for unit in set(profileData.Unit.tolist()):
                    element_unit_data = profileData[profileData["Unit"] == unit]
                    values = ";".join(element_unit_data.ValuesX.to_list())
                    values = [float(v) for v in values.split(";")]
                    elements[id][unit] = values

            dict_of_df = {k: pd.DataFrame(v) for k, v in elements.items()}
            elements = pd.concat(dict_of_df, axis=1)
            elements.index = pd.Timestamp(f'{year}-01-01')+pd.to_timedelta(elements.index * self.freq[dTtype], unit="m")
            data[dTtype] = elements

        for eType, eData in data.items():
            for elm_name, unit_id in eData.columns:
                values = eData[elm_name][unit_id]
                device = self.get_device(eType, elm_name)
                #print(eType, elm_name, device)
                if device:
                    cName = CLASS_MAP[device.DeviceType]
                    devices = {
                        f"{cName}.{device.DeviceNumber}": device
                    }
                    self.Profiles[f"{cName}/{device.DeviceNumber}/{UNITS_MAP[int(unit_id)]}"] = Profile(
                        self.sim_instance,
                        values,
                        devices,
                        self.solver,
                        None,
                        self.logger,
                        **{
                            "type": f"{cName}",
                            "name": f"{device.DeviceNumber}",
                            "unit": f"{UNITS_MAP[int(unit_id)]}",
                        }
                    )

    def update(self):
        results = {}
        for profileaName, profileObj in self.Profiles.items():
            result = profileObj.update()
            results[profileaName] = result
        return results

    def get_device(self, dTtype, id):
        devType = getattr(self.sim_instance.enums.DeviceType, dTtype)
        devices = self.sim_instance.study.ListDevices(devType)
        for device in devices:
            if id == device.DeviceNumber:
                return device
        return None

class Profile(BaseProfile):
    def __init__(self, sim_instance, dataset, devices, solver, mapping_dict, logger, **kwargs):
        super(Profile, self).__init__(sim_instance, dataset, devices, solver, mapping_dict, logger, **kwargs)
        name = list(devices.keys())[0]
        self.valueSettings = {name: DEFAULT_PROFILE_SETTINGS}
        self.attrs = {
            "sTime": self.dataset.index.min(),
            "eTime": self.dataset.index.max(),
            "units": kwargs['unit'].encode(),
            "info": "",
            "mean": self.dataset.mean(),
            "min": self.dataset.min(),
            "max": self.dataset.max(),
            "npts": len(self.dataset),
            "resTime": 3600,

        }
        self.logger.info(f"Profile '{kwargs['type']}' created and connect tp ppty '{kwargs['unit']}' of device'{kwargs['name']}' ")

    def update_profile_settings(self):
        pass

    def update(self):
        self.Time = copy.deepcopy(self.solver.GetDateTime())
        if self.Time < self.dataset.index.min() or self.Time > self.dataset.index.max():
            value = 0
        else:
            value = self.dataset[self.Time]
        value = self.write(value, value)
        return value


from cymepy.export_manager.base_definations import ExportManager
import os

class Writer(ExportManager):
    def __init__(self, sim_instance, solver, options, logger, **kwargs):
        super(Writer, self).__init__(sim_instance, solver, options, logger, **kwargs)
        self.results = ["Timestamp,Class,Element,Property,Value\n"]

        self.path = os.path.join(
            self.settings["project"]['project_path'],
            'exports',
        )
        self.results = []
        self.step=0
        return

    def update(self):
        results = super().update()
        self.results.append(results)
        return

    def export(self):
        FileHandles = {}
        for results in self.results:
            for timestamp, cInfo in results.items():
                for cName, eInfo in cInfo.items():
                    if cName not in FileHandles:
                        FileHandles[cName] = open(
                            os.path.join(
                                self.path, f"{cName}.csv"
                            ),
                            "w"
                        )
                        self.writeHeader = True
                    else:
                        self.writeHeader = False
                    headers = ["Timestamp"]
                    values = [timestamp]
                    for eName, pptyInfo in eInfo.items():
                        for ppty, value in pptyInfo.items():
                            headers.append(f"{eName}.{ppty}")
                            values.append(value)
                    if self.writeHeader:
                        headers = ",".join(headers)
                        FileHandles[cName].write(headers + "\n")
                    values = ",".join(values)
                    FileHandles[cName].write(values + "\n")

        for cNama, handler in FileHandles.items():
            handler.flush()
            handler.close()

